package snmp

import (
	"fmt"
	"log"
	"time"
  "encoding/json"
  "strconv"
  "math"

	"smq/config"
	"smq/mqtt"

	"github.com/gosnmp/gosnmp"
)

type variable_t struct {
  Label string `json:"label"`
  Variable string `json:"variable"`
  Value string `json:"value"`
  Unit string `json:"unit"`
  Status string `json:"status"`
}

type message_t struct {
  Device string `json:"device"`
  Label string `json:"label"`
  Variables map[string]variable_t `json:"variables"`
}

type lastValue_t struct {
  Timestamp time.Time
  Value string
}

var lastValues map[string]lastValue_t = make(map[string]lastValue_t)

func calculateDifference(key string, newValue string) (string, error) {
  currentTime := time.Now()
  lvv, ok := lastValues[key]
  if ok {
    log.Printf("lvv for %s exists: %s", key, lvv.Value)
    oldValueI, err1 := strconv.Atoi(lvv.Value)
    if err1 != nil {
      return "", fmt.Errorf("failed to convert lastValue: %s, %v", lvv.Value, err1)
    }
    newValueI, err2 := strconv.Atoi(newValue)
    if err2 != nil {
      return "", fmt.Errorf("failed to convert newValue: %s, %v", newValue, err2)
    }
    diffValueI := newValueI - oldValueI
    diffTime := int(math.Round(currentTime.Sub(lvv.Timestamp).Seconds()))
    diffValuePerSecond := diffValueI / diffTime
    log.Printf("Diff: %d, Duration: %d, Diff per second: %d", diffValueI, diffTime, diffValuePerSecond)
    lastValues[key] = lastValue_t {
      Timestamp: currentTime,
      Value: newValue,
    }
    return strconv.Itoa(diffValuePerSecond), nil
  } else {
    log.Printf("create lvv for %s", key)
    lastValues[key] = lastValue_t {
      Timestamp: currentTime,
      Value: newValue,
    }
    return "0", nil
  }
  return "0", nil
}

func Start() {
	for {
    for _, endpoint := range config.Config.SNMPEndpoints {
      log.Println("Polling endpoint " + endpoint.Endpoint)

      snmp := gosnmp.GoSNMP{}

      snmp.Target = endpoint.Endpoint
      snmp.Port = 161
      snmp.Version = gosnmp.Version2c
      snmp.Community = endpoint.Community

      snmp.Timeout = time.Duration(5 * time.Second)
      err := snmp.Connect()
      if err != nil {
        log.Fatal("SNMP Connect error\n")
      }

      oids := []string{}

      for _, oidTopic := range endpoint.OIDTopics {
        oids = append(oids, oidTopic.OID)
      }

      result, err := snmp.Get(oids)
      if err != nil {
        log.Printf("error in Get: %s", err)
      } else {
        message := message_t {
          Device: endpoint.Endpoint,
          Label: endpoint.Label,
          Variables: make(map[string]variable_t),
        }
        

        for _, variable := range result.Variables {
          for _, oidTopic := range endpoint.OIDTopics {
            if oidTopic.OID == variable.Name {
              convertedValue := ""

              switch variable.Type {
              case gosnmp.OctetString:
                convertedValue = string(variable.Value.([]byte))
              default:
                convertedValue = fmt.Sprintf("%d", gosnmp.ToBigInt(variable.Value))
              }

              if oidTopic.Diff == "true" {
                log.Println("Calculate difference to last value")
                key := endpoint.Endpoint + ":" + oidTopic.OID
                diff, err := calculateDifference(key, convertedValue)
                if err != nil {
                  log.Printf("Error when building difference: %v", err)
                  convertedValue = "-1"
                } else if diff < 0 {
                  convertedValue = "-2"
                } else {
                  convertedValue = diff 
                }
              }

              log.Printf("%s = %s", oidTopic.OID, convertedValue)
              v := variable_t {
                Label: oidTopic.Label,
                Variable: oidTopic.OID,
                Value: convertedValue,
                Unit: "",
                Status: "",
              }
              message.Variables[oidTopic.Label] = v

              // build topic and payload and push into the mqtt.OutputChannel
            }
          }
        }

        j, err := json.Marshal(message)
        if err != nil {
          log.Printf("Unable to marshal message, it is lost: %s, %v", message, err)
        } else {
          mqtt.Publish(j)
        }
      }
      snmp.Conn.Close()
    }

		time.Sleep(time.Duration(config.Config.Interval) * time.Second)
	}
}

func Stop() {
}
